
#include "book_utils.h"

extern int verbose;

Display*
ora_init_pex( argv, pexinfo )
    char		*argv[];
    PEXExtensionInfo	**pexinfo;
{
    char	*dpy_name, err[80];
    Display	*dpy = (Display *)NULL;
    
    if ( argv[1] && strcmp( "-display", argv[1] ) == 0 )
	dpy_name = XDisplayName( argv[2] );
    else	    
        dpy_name = XDisplayName( NULL );
    
    if ( !(dpy = XOpenDisplay( dpy_name )) ){
	fprintf ( stderr, "Could not open display %s\n", dpy_name );
    
    } else if ( PEXInitialize( dpy, pexinfo, 80, err ) != 0 ) {
	fprintf( stderr, "No PEX extension on %s.\n", dpy_name );
	fprintf( stderr, "%s\n", err );
	XCloseDisplay( dpy );
	dpy = (Display *)NULL;
    }
    
    return dpy;
}

static void
find_best_visual_5_0( dpy, chosen_vis )
    Display		*dpy;
    XVisualInfo		*chosen_vis;
{
    int			i, size, num_vis, cmap_size;
    XVisualInfo		vis_templ, *vis_list;
    
    /* Get all the visuals for the screen. */
    vis_templ.screen = DefaultScreen( dpy );
    vis_list = XGetVisualInfo( dpy, VisualScreenMask, &vis_templ,
			      &num_vis );
    
    /* Determine the best visual available.  The best one is the */
    /* one with the most colors and highest capabilities. */
    cmap_size = 0;
    chosen_vis->class = -1;
    for ( i = 0; i < num_vis; i++ ) {
	/* Determine the number of colors available. */
	switch ( vis_list[i].class ) {
	case TrueColor:
	case DirectColor:
	    size = vis_list[i].red_mask | vis_list[i].green_mask
		| vis_list[i].blue_mask + 1;
	    break;
	default:
	    size = vis_list[i].colormap_size;
	}
	
	/* Choose this one if it supports more colors or is a higher */
	/* class, but favor TrueColor over DirectColor. */
	if ( size >= cmap_size ) {
	    switch ( vis_list[i].class ) {
	    case TrueColor:
		/* Choose TrueColor over anything else. */
		*chosen_vis = vis_list[i];
		cmap_size = size;
		break;
	    case DirectColor:
		/* Don't choose DirectColor over TrueColor. */
	        if ( chosen_vis->class != TrueColor ) {
		    *chosen_vis = vis_list[i];
		    cmap_size = size;
		}
		break;
	    default:
	        /* Choose the highest class. */
		if ( vis_list[i].class >= chosen_vis->class ) {
		    *chosen_vis = vis_list[i];
		    cmap_size = size;
		}
		break;
	    }
	}
    }
}

static int
find_best_visual_5_1( dpy, chosen_vis )
    Display		*dpy;
    XVisualInfo		*chosen_vis;
{
    int                 i, depth = 0, size, cmap_size = 0;
    int                 chosen_class;
    unsigned long       num_targets, max_targets = 100;
    PEXRenderingTarget *targets, chosen_target;
    Visual             *visual = (Visual *)NULL;
    int                 type = PEXWindowDrawable;
    int                 double_buffer_not_supported = 0;

    /* Get all visuals the PEX extension on this server supports. */
    PEXMatchRenderingTargets( dpy,
			      RootWindow(dpy,DefaultScreen(dpy)),
			      depth, PEXAnyDrawable, visual,
			      max_targets, &num_targets, &targets );
    
    /* Find the one with the most colors and capabilities. */
    chosen_class = -1;

    for ( i = 0; i < num_targets; i++ ) {

    if( targets[i].type == PEXPixmapDrawable )
        type = PEXPixmapDrawable;

	visual = targets[i].visual;
	
	/* Determine the number of colors available. */
	switch ( visual->class ) {
	case TrueColor:
	case DirectColor:
	    size = visual->red_mask | visual->green_mask
		| visual->blue_mask + 1;
	    break;
	default:
	    size = visual->map_entries;
	}
	
	/* Choose this one if it supports more colors or is a higher */
	/* class, but favor TrueColor over DirectColor. */
    if( double_buffer_supported( dpy, visual, targets[i].depth ) ) {
        double_buffer_not_supported = 1;
        if ( (size >= cmap_size) || (visual->class > chosen_class) ) {
            switch ( visual->class ) {
            case TrueColor:
                /* Choose TrueColor over anything else. */
                chosen_target = targets[i];
                chosen_class = visual->class;
                cmap_size = size;
                break;
            case DirectColor:
                if ( targets[i].depth == 8 ) break;
                /* Don't choose DirectColor over TrueColor. */
                if ( chosen_class != TrueColor ) {
                    chosen_target = targets[i];
                    chosen_class = visual->class;
                    cmap_size = size;
                }
                break;
            default:
                /* Choose the highest class. */
                if ( visual->class >= chosen_class ) {
                    chosen_target = targets[i];
                    chosen_class = visual->class;
                    cmap_size = size;
                }
                break;
            }
        }
    }
    }

    if( !double_buffer_not_supported )
    for ( i = 0; i < num_targets; i++ ) {

    if( targets[i].type == PEXPixmapDrawable )
        type = PEXPixmapDrawable;

    visual = targets[i].visual;

    /* Determine the number of colors available. */
    switch ( visual->class ) {
    case TrueColor:
    case DirectColor:
        size = visual->red_mask | visual->green_mask
            | visual->blue_mask + 1;
        break;
    default:
        size = visual->map_entries;
    }

    /* Choose this one if it supports more colors or is a higher */
    /* class, but favor TrueColor over DirectColor. */
    if ( (size >= cmap_size) || (visual->class > chosen_class) ) {
        switch ( visual->class ) {
        case TrueColor:
            /* Choose TrueColor over anything else. */
            chosen_target = targets[i];
            chosen_class = visual->class;
            cmap_size = size;
            break;
        case DirectColor:
            if ( targets[i].depth == 8 ) break;
            /* Don't choose DirectColor over TrueColor. */
            if ( chosen_class != TrueColor ) {
                chosen_target = targets[i];
                chosen_class = visual->class;
                cmap_size = size;
            }
            break;
        default:
            /* Choose the highest class. */
            if ( visual->class >= chosen_class ) {
                chosen_target = targets[i];
                chosen_class = visual->class;
                cmap_size = size;
            }
            break;
        }
    }
    }

    if ( num_targets > 0 ) {
	chosen_vis->visual = chosen_target.visual;
	chosen_vis->visualid = chosen_target.visual->visualid;
	chosen_vis->class = chosen_target.visual->class;
	chosen_vis->red_mask = chosen_target.visual->red_mask;
	chosen_vis->green_mask = chosen_target.visual->green_mask;
	chosen_vis->blue_mask = chosen_target.visual->blue_mask;
	chosen_vis->colormap_size = chosen_target.visual->map_entries;
	chosen_vis->bits_per_rgb = chosen_target.visual->bits_per_rgb;
	chosen_vis->screen = DefaultScreen(dpy);
	chosen_vis->depth = chosen_target.depth;
        XFree( targets );
    }
    if( verbose )
    switch ( chosen_vis->class ) {
        case DirectColor:
            fprintf( stderr, "Chose:\t%d-bit DirectColor visual\n",
                chosen_target.depth );
            break;
        case TrueColor:
            fprintf( stderr, "Chose:\t%d-bit TrueColor visual\n",
                chosen_target.depth );
            break;
        default:
            fprintf( stderr, "Using:\t%d-bit PseudoColor visual\n",
                chosen_target.depth );
            break;
    }
    return( type );
}

int
ora_find_best_visual( dpy, vis_info )
Display     *dpy;
XVisualInfo *vis_info;
{
    PEXExtensionInfo    *pex_info;

    pex_info = PEXGetExtensionInfo( dpy );

    if ( VERSION_5_0( pex_info ) ) {
        find_best_visual_5_0( dpy, vis_info );
        return( PEXWindowDrawable );
    }
    else if ( VERSION_5_1( pex_info ) )
        return( find_best_visual_5_1( dpy, vis_info ) );
}

int
ora_get_standard_colormap( dpy, vis_info, cmap_info )
    Display		*dpy;
    XVisualInfo		*vis_info;
    XStandardColormap	*cmap_info;
{
    XStandardColormap	*std_cmaps;
    Atom		property;
    int			i, num_cmaps;
    
    switch ( vis_info->class ) {
    case TrueColor:
    case StaticColor:
	    property = XA_RGB_BEST_MAP;
        if( verbose )
	    fprintf( stderr, "Set XA_RGB_BEST_MAP property\n" );
	    break;
    case PseudoColor:
    case DirectColor:
    default:
	    property = XA_RGB_DEFAULT_MAP;
        if( verbose )
	    fprintf( stderr, "Set XA_RGB_DEFAULT_MAP property\n" );
	    break;
    case StaticGray:
    case GrayScale:
	    property = XA_RGB_GRAY_MAP;
        if( verbose )
	    fprintf( stderr, "Set XA_RGB_GRAY_MAP property\n" );
	    break;
    }
cmap:
    /* Ensure that the property is defined. */
    if ( XmuLookupStandardColormap( dpy, vis_info->screen,
		vis_info->visualid, vis_info->depth, property,
		False, True ) ) {
	
	/* Get the standard colormap properties. */
	if ( XGetRGBColormaps( dpy,
	      RootWindow( dpy, DefaultScreen( dpy ) ),
	      &std_cmaps, &num_cmaps, property ) ) {
	
	    /* Find the properties for the specified visual. */
	    for ( i = 0; i < num_cmaps; i++ ) {
		if ( vis_info->visualid == std_cmaps[i].visualid ) {
		    *cmap_info = std_cmaps[i];
		    return 1;
		}
	    }
	    XFree( std_cmaps );
	}
    }
    
    if ( property == XA_RGB_DEFAULT_MAP ) {
        property = XA_RGB_BEST_MAP;
        if( verbose )
        fprintf( stderr, "Set XA_RGB_BEST_MAP property\n" );
        goto cmap;
    }

    return 0;
}

void
ora_set_stdcmap_approx( vis_info, cmap_info, capx_info )
    XVisualInfo		*vis_info;
    XStandardColormap	*cmap_info;
    PEXColorApproxEntry	*capx_info;
{
    switch ( vis_info->class ) {
    case DirectColor:
    case TrueColor:
    case PseudoColor:
    case StaticColor:
    default:
	    capx_info->type = PEXColorSpace;
	    capx_info->model = PEXColorApproxRGB;
	    capx_info->dither = PEXOn;
	    capx_info->base_pixel = cmap_info->base_pixel;
	    capx_info->max1 = cmap_info->red_max;
	    capx_info->max2 = cmap_info->green_max;
	    capx_info->max3 = cmap_info->blue_max;
	    capx_info->weight1 = 0; /* not used by PEXColorSpace */
	    capx_info->weight2 = 0; /* not used by PEXColorSpace */
	    capx_info->weight3 = 0; /* not used by PEXColorSpace */
	    capx_info->mult1 = cmap_info->red_mult;
	    capx_info->mult2 = cmap_info->green_mult;
	    capx_info->mult3 = cmap_info->blue_mult;
	    break;
    case GrayScale:
    case StaticGray:
	    capx_info->type = PEXColorRange;
	    capx_info->model = PEXColorApproxRGB;
	    capx_info->dither = PEXOn;
	    capx_info->base_pixel = cmap_info->base_pixel;
	    capx_info->max1 = cmap_info->red_max;
	    capx_info->max2 = 0; /* not used by PEXColorRange */
	    capx_info->max3 = 0; /* not used by PEXColorRange */
	    /* Give the weights the NTSC intensity coefficients. */
	    capx_info->weight1 = 0.299; 
	    capx_info->weight2 = 0.587;
	    capx_info->weight3 = 0.114;
	    capx_info->mult1 = cmap_info->red_mult;
	    capx_info->mult2 = 0;
	    capx_info->mult3 = 0;
	    break;
    }
    if( verbose ) {
	fprintf( stderr, "Color Approximation:\n" );
	fprintf( stderr, "\tmax1\t= %d\n", capx_info->max1 );
	fprintf( stderr, "\tmax2\t= %d\n", capx_info->max2 );
	fprintf( stderr, "\tmax3\t= %d\n", capx_info->max3 );
	fprintf( stderr, "\tmult1\t= %d\n", capx_info->mult1 );
	fprintf( stderr, "\tmult2\t= %d\n", capx_info->mult2 );
	fprintf( stderr, "\tmult3\t= %d\n", capx_info->mult3 );
	fprintf( stderr, "\tweight1\t= %d\n", capx_info->weight1 );
	fprintf( stderr, "\tweight2\t= %d\n", capx_info->weight2 );
	fprintf( stderr, "\tweight3\t= %d\n", capx_info->weight3 );
	fprintf( stderr, "\tdither\t= %d\n", capx_info->dither );
	fprintf( stderr, "\tbase_pixel = %d\n", capx_info->base_pixel );
    }
}

Window
ora_create_window( dpy, parent, vis_info, cmap_info, geom, title, iconTitle )
    Display		*dpy;
    Window      parent;
    XVisualInfo		*vis_info;
    XStandardColormap	*cmap_info;
    char		*geom;
    char        *title;
    char        *iconTitle;
#if( 0 ) 
    XColor		*bkgd_colr;
#endif
{
    Window			window;
    XSetWindowAttributes	wattrs;
    int				x = 0, y = 0;
    unsigned int		width = 300, height = 300;
    XEvent      event;
    XSizeHints  size_hints;
#if( 0 ) 
    /* Determine the pixel value for the background. */
    if ( XAllocColor( dpy, cmap_info->colormap, bkgd_colr ) == 0 )
	/* Colormap is read/write and no more entries left.  So use */
	/* the colormap's base pixel, which will probably be black. */
	bkgd_colr->pixel = cmap_info->base_pixel;

    /* Create the output window. */
    wattrs.background_pixel = bkgd_colr->pixel;
    wattrs.border_pixel = bkgd_colr->pixel;
#endif
    wattrs.background_pixel = 0;
    wattrs.border_pixel = 0;
    wattrs.backing_store = NotUseful;
    wattrs.event_mask = ExposureMask;
    wattrs.colormap = cmap_info->colormap;
    XParseGeometry( geom, &x, &y, &width, &height );
    window = XCreateWindow( dpy, parent, x, y, width, height, 1,
	    vis_info->depth, InputOutput, vis_info->visual,
	    CWBackPixel | CWBorderPixel | CWBackingStore | CWEventMask | CWColormap,
	    &wattrs );

    /* Finish the setup. */
    size_hints.flags = USSize;
    size_hints.width = width;
    size_hints.height = height;
    XSetStandardProperties( dpy, window, title, iconTitle,
                            None, NULL, 0, &size_hints );

    XSelectInput( dpy, window, (long)ExposureMask );
    XMapWindow( dpy, window );
    XSync( dpy, False );
    XNextEvent( dpy, &event );
    XSelectInput( dpy, window, (long)0 );

    return window;
}

static void
load_all_available_fonts( dpy, window, table )
    Display		*dpy;
    Window		window;
    PEXLookupTable	table;
{
    int			i, max;
    unsigned long	num_fonts;
    char		**font_names;
    PEXFont		*font_ids;
    PEXTextFontEntry	*entries;
    PEXTableInfo	table_info;
    PEXExtensionInfo	*pex_info;

    /* The X Consortium's sample server crashes when attempting to */
    /* load fonts, and has them pre-loaded anyway.  So return if */
    /* that's the server we're dealing with. */
    pex_info = PEXGetExtensionInfo( dpy );
    if ( strstr( pex_info->vendor_name, "Sample Implementation" ) )
	return;

    /* Determine the font table limit. */
    if ( !PEXGetTableInfo( dpy, window, PEXLUTTextFont, &table_info ) )
        return;	/* bad display pointer or window */
    max = table_info.definable_entries;
    
    /* Allocate space for the font ids and table entries. */
    font_ids = (PEXFont *) malloc( max * sizeof(PEXFont) );
    entries = (PEXTextFontEntry *)
		malloc( max * sizeof(PEXTextFontEntry) );
    
    /* Get the font names, load the fonts, and store in the table. */
    if ( font_names = PEXListFonts( dpy, "*", max, &num_fonts ) ) {
	for ( i = 0; i < num_fonts; i++ ) {
	    font_ids[i] = PEXLoadFont( dpy, font_names[i] );
	    entries[i].count = 1;
	    entries[i].fonts = &font_ids[i];
	}
    
	PEXSetTableEntries( dpy, table, 1, num_fonts,
			    PEXLUTTextFont, (PEXPointer) entries );
	PEXFreeFontNames(  num_fonts, font_names );
    }
    free( font_ids );
    free( entries );
}

int
ora_enum_type_supported( dpy, window, enum_type, value )
    Display	*dpy;
    Window	window;
    int		enum_type;
    int		value;
{
    int				enum_types[1], i, found = 0;
    unsigned long		*num_values;
    PEXEnumTypeDesc		*values;
    
    enum_types[0] = enum_type;
    if ( PEXGetEnumTypeInfo( dpy, window, 1, enum_types,
			     PEXETIndex, &num_values, &values ) ) {
	for ( i = 0; i < *num_values; i++ ) {
	    if ( values[i].index == value ) {
		found = 1; /* found it! */
		break; /* out of for loop */
	    }
	}
	PEXFreeEnumInfo( 1, num_values, values );
    }		
    
    return found;	    
}
    
PEXRenderer
ora_setup_renderer( dpy, window, capx_info, attrs )
    Display			*dpy;
    Window			window;
    PEXColorApproxEntry		*capx_info;
    PEXRendererAttributes	*attrs;
{
    unsigned long		mask = 0;
    PEXRenderer			renderer;
    PEXExtensionInfo		*pex_info;
    
    /* Create all the bundle tables. */
    mask |= PEXRALineBundle;
    attrs->line_bundle =
	PEXCreateLookupTable( dpy, window, PEXLUTLineBundle );
    
    mask |= PEXRAMarkerBundle;
    attrs->marker_bundle =
	PEXCreateLookupTable( dpy, window, PEXLUTMarkerBundle );
    
    mask |= PEXRATextBundle;
    attrs->text_bundle =
	PEXCreateLookupTable( dpy, window, PEXLUTTextBundle );
    
    mask |= PEXRAInteriorBundle;
    attrs->interior_bundle =
	PEXCreateLookupTable( dpy, window, PEXLUTInteriorBundle );
    
    mask |= PEXRAEdgeBundle;
    attrs->edge_bundle =
	PEXCreateLookupTable( dpy, window, PEXLUTEdgeBundle );
    
    /* Conditionally create the pattern table. */
    if ( ora_enum_type_supported( dpy, window, PEXETInteriorStyle,
				  PEXInteriorStylePattern ) ) {
	mask |= PEXRAPatternTable;
	attrs->pattern_table =
	    PEXCreateLookupTable( dpy, window, PEXLUTPattern );
    }	    
    
    /* Create a font table. */
    mask |= PEXRATextFontTable;
    attrs->text_font_table =
	PEXCreateLookupTable( dpy, window, PEXLUTTextFont );
    load_all_available_fonts( dpy, window, attrs->text_font_table );
    
    /* Create a color table. */
    mask |= PEXRAColorTable;
    attrs->color_table =
	PEXCreateLookupTable( dpy, window, PEXLUTColor );
    
    /* Create a view table. */
    mask |= PEXRAViewTable;
    attrs->view_table =
	PEXCreateLookupTable( dpy, window, PEXLUTView );

    /* Create a light table. */
    mask |= PEXRALightTable;
    attrs->light_table =
	PEXCreateLookupTable( dpy, window, PEXLUTLight );
    
    /* Create a depth cue table. */
    mask |= PEXRADepthCueTable;
    attrs->depth_cue_table =
	PEXCreateLookupTable( dpy, window, PEXLUTDepthCue );

    /* Create a color approximation table and set the default */
    /* entry, entry 0, to the colormap approximation specified. */
    mask |= PEXRAColorApproxTable;
    attrs->color_approx_table =
	PEXCreateLookupTable( dpy, window, PEXLUTColorApprox );
    PEXSetTableEntries( dpy, attrs->color_approx_table, 0, 1,
		        PEXLUTColorApprox, capx_info );

    /* Create the highlighting, invisibility, and pick name sets. */
    mask |= PEXRAHighlightIncl | PEXRAHighlightExcl;
    attrs->highlight_incl = PEXCreateNameSet( dpy );
    attrs->highlight_excl = PEXCreateNameSet( dpy );
    
    mask |= PEXRAInvisibilityIncl | PEXRAInvisibilityExcl;
    attrs->invisibility_incl = PEXCreateNameSet( dpy );
    attrs->invisibility_excl = PEXCreateNameSet( dpy );
    
    mask |= PEXRAPickIncl | PEXRAPickExcl;
    attrs->pick_incl = PEXCreateNameSet( dpy );
    attrs->pick_excl = PEXCreateNameSet( dpy );

    mask |= PEXRAHLHSRMode;
    attrs->hlhsr_mode = PEXHLHSRZBuffer;

    pex_info = PEXGetExtensionInfo( dpy );

    if ( VERSION_5_1( pex_info ) ) {
        mask |= PEXRAClearImage;
        attrs->clear_image = True;

        mask |= PEXRAClearZ;
        attrs->clear_z = True;

        mask |= PEXRABackgroundColor;
        attrs->background_color.type = PEXColorTypeRGB;
        attrs->background_color.value.rgb.red = 0.0;
        attrs->background_color.value.rgb.green = 0.0;
        attrs->background_color.value.rgb.blue = 0.0;
    }
    
    /* Create the renderer */
    renderer = PEXCreateRenderer( dpy, window, mask, attrs );
    
    return renderer;
}

int
ora_get_colormap_and_visual( dpy, property, vis_info, cmap_info )
    Display		*dpy;
    Atom		property;
    XVisualInfo		*vis_info;
    XStandardColormap	*cmap_info;
{
    XStandardColormap		*std_cmaps;
    XVisualInfo			vis_templ, *visual;
    int			        num_cmaps, num_vis, i, cmap_size, size;
    char		        buf[256];
    Status			st;

    
    /* Get the colormap properties of the requested maps. */
    st = XGetRGBColormaps( dpy,
			  RootWindow( dpy, DefaultScreen( dpy ) ),
			  &std_cmaps, &num_cmaps, property );
    if ( st == 0 ) {
	/* The standard colormaps are not loaded.  Try to load. */
	XmuAllStandardColormaps( dpy );

	
	/* Try again to get the properties. */
	st = XGetRGBColormaps( dpy,
			      RootWindow( dpy, DefaultScreen( dpy ) ),
			      &std_cmaps, &num_cmaps, property );
	if ( st == 0 ) {
	    fprintf( stderr, "Utils: Sorry, the standard colormap is not" );
	    fprintf( stderr, " set and cannot be loaded.\n" );
	    return 0;
	}
    }
    
    /* Determine the best visual available.  The best one is the */
    /* one with the largest colormap size (the most colors). */
    cmap_size = 0;
    for ( i = 0; i < num_cmaps; i++ ) {
	/* Get the visual info for the indicated visual id. */
	vis_templ.visualid = std_cmaps[i].visualid;
	visual = XGetVisualInfo( dpy, VisualIDMask, &vis_templ,
				&num_vis );
	
	/* See if the visual's got the largest color table. */
	switch ( visual->class ) {
	case TrueColor:
	case DirectColor:
	    size = visual->red_mask | visual->green_mask
		| visual->blue_mask + 1;
	    break;
	default:
	    size = visual->colormap_size;
	}
	
	if ( size > cmap_size ) {
	    *vis_info = *visual;
	    *cmap_info = std_cmaps[i];
	    cmap_size = visual->colormap_size;
	}
    }
    
    XFree( std_cmaps );
    XFree( visual );
    return 1;
}

int
ora_create_colormap( dpy, vis_info, cmap_info, capx_info )
    Display		*dpy;
    XVisualInfo		*vis_info;
    XStandardColormap	*cmap_info;
    PEXColorApproxEntry	*capx_info;
{
    int			st = 0, n;
    
    switch ( vis_info->class ) {
    case DirectColor:
        /* Create the largest possible equal-length ramps. */
	st = ora_create_direct_map( dpy, vis_info, cmap_info,
				    capx_info );
    if( verbose )
	fprintf( stderr, "Created Direct Colormap\n" );
	break;
    case PseudoColor:
        /* Create the largest NxNxN color sampling. */
	n = (int)pow( (double)vis_info->colormap_size - 1, 1.0/3.0 );
	st = ora_create_pseudo_map( dpy, vis_info, n, n, n,
				    cmap_info, capx_info );
    if( verbose )
	fprintf( stderr, "Created Pseudo Colormap\n" );
	break;
    case GrayScale:
	/* Create a GrayScale colormap with max number of grays. */
	/* (but leave one empty spot for the background color.) */
	st = ora_create_gray_map( dpy, vis_info,
				  vis_info->colormap_size - 1,
				  cmap_info, capx_info );
    if( verbose )
	fprintf( stderr, "Created Gray Colormap\n" );
	break;
    case TrueColor:
    case StaticColor:
    case StaticGray:
	st = ora_create_read_only_map( dpy, vis_info, cmap_info,
				       capx_info );
    if( verbose )
	fprintf( stderr, "Created Read Only (True) Colormap\n" );
	break;
    }
    
    return st;
}    

int
ora_create_direct_map( dpy, vis_info, cmap_info, capx_info )
    Display		*dpy;
    XVisualInfo		*vis_info;
    XStandardColormap	*cmap_info;
    PEXColorApproxEntry	*capx_info;
{
    int			st, red_planes, green_planes, blue_planes;
    unsigned long	i;
    unsigned long	num_reds, num_greens, num_blues, num_colors;
    unsigned long	rshift, gshift, bshift;
    unsigned long	rmask, gmask, bmask;
    XColor		*colors, *color;
    
    /* Create the colormap. */
    cmap_info->visualid = vis_info->visualid;
    cmap_info->colormap = XCreateColormap( dpy,
	RootWindow( dpy, DefaultScreen( dpy )),
	vis_info->visual, AllocNone );

    /* Determine the number of red, green, and blue planes and the */
    /* maximum possible number of color values for each. */
    red_planes = 0; num_reds = 1;
    for( i = vis_info->red_mask; i; i = i >> 1 )
	if ( i & 1 ) {
	    ++red_planes;
	    num_reds *= 2;
	}
    cmap_info->red_max = num_reds - 1;
	
    green_planes = 0; num_greens = 1;
    for( i = vis_info->green_mask; i; i = i >> 1 )
	if ( i & 1 ) {
	    ++green_planes;
	    num_greens *= 2;
	}
    cmap_info->green_max = num_greens - 1;
	
    blue_planes = 0; num_blues = 1;
    for( i = vis_info->blue_mask; i; i = i >> 1 )
	if ( i & 1 ) {
	    ++blue_planes;
	    num_blues *= 2;
	}
    cmap_info->blue_max = num_blues - 1;
    
    /* Allocate the planes. */
    st = XAllocColorPlanes( dpy, cmap_info->colormap, True,
			    &cmap_info->base_pixel, 1,
			    red_planes, green_planes, blue_planes,
			    &rmask, &gmask, &bmask );
    if ( st == 0 )
	return 0;

    /* Determine the red, green, and blue multipliers by finding */
    /* the first bit set in each mask. */
    for ( rshift = 0; !(rmask & (1 << rshift)); rshift++ )
	;
    cmap_info->red_mult = 1 << rshift;
    
    for ( gshift = 0; !(gmask & (1 << gshift)); gshift++ )
        ;
    cmap_info->green_mult = 1 << gshift;
    
    for ( bshift = 0; !(bmask & (1 << bshift)); bshift++ )
	;
    cmap_info->blue_mult = 1 << bshift;

    /* Store the colors in the colormap. */
    num_colors = vis_info->colormap_size;
    colors = (XColor *) malloc( num_colors * sizeof(XColor) );
    for ( i = 0, color = colors; i < num_colors; i++, color++ ) {
	color->flags = 0;
	color->pixel = cmap_info->base_pixel;
	if ( i < cmap_info->red_max ) {
	    color->flags |= DoRed ;
	    color->pixel |= (i << rshift);
	    color->red = i * 65535 / cmap_info->red_max;
	}
	if ( i < cmap_info->green_max ) {
	    color->flags |= DoGreen ;
	    color->pixel |= (i << gshift);
	    color->green = i * 65535 / cmap_info->green_max;
	}
	if ( i < cmap_info->blue_max ) {
	    color->flags |= DoBlue;
	    color->pixel |= (i << bshift);
	    color->blue = i * 65535 / cmap_info->blue_max;
	}
    }
    XStoreColors( dpy, cmap_info->colormap, colors, num_colors );
    free( colors );
    
    /* Fill in the color approximation information. */
    capx_info->type = PEXColorSpace;
    capx_info->model = PEXColorApproxRGB;
    capx_info->dither = PEXOn;
    capx_info->base_pixel = cmap_info->base_pixel;
    capx_info->max1 = cmap_info->red_max;
    capx_info->max2 = cmap_info->green_max;
    capx_info->max3 = cmap_info->blue_max;
    capx_info->weight1 = 0.0; /* not used by PEXColorSpace */
    capx_info->weight2 = 0.0; /* not used by PEXColorSpace */
    capx_info->weight3 = 0.0; /* not used by PEXColorSpace */
    capx_info->mult1 = cmap_info->red_mult;
    capx_info->mult2 = cmap_info->green_mult;
    capx_info->mult3 = cmap_info->blue_mult;
    
    return 1;
}

int
ora_create_pseudo_map( dpy, vis_info, nr, ng, nb, cmap_info, capx_info )
    Display		*dpy;
    XVisualInfo		*vis_info;
    int			nr, ng, nb;
    XStandardColormap	*cmap_info;
    PEXColorApproxEntry	*capx_info;
{
    int			num_colors, st, i, j, k, p;
    unsigned long	*pixels;
    XColor		*colors, *color;
    
    /* Create the colormap and fill in the standard cmap info. */
    cmap_info->colormap = XCreateColormap( dpy,
	RootWindow( dpy, DefaultScreen( dpy )),
	vis_info->visual, AllocNone );
    cmap_info->visualid = vis_info->visualid;
    cmap_info->blue_max = nb - 1;	
    cmap_info->blue_mult = 1;
    cmap_info->green_max = ng - 1;	
    cmap_info->green_mult = nb;
    cmap_info->red_max = nr - 1;	
    cmap_info->red_mult = nb * ng;

    num_colors = nr * ng * nb;
    pixels = (unsigned long*) malloc( num_colors * sizeof(*pixels) );
    st = XAllocColorCells( dpy, cmap_info->colormap, True,
			  (unsigned long *)NULL, 0, pixels, num_colors );
    if ( st == 0 ) {
	free( pixels );
	return 0;
    }
	
    cmap_info->base_pixel = pixels[0];			  
    free( pixels );

    colors = (XColor *) malloc( num_colors * sizeof(*colors) );
    p = cmap_info->base_pixel;
    for ( i = 0, color = colors; i < nr; i++ ) {
	for ( j = 0; j < ng; j++ ) {
	    for ( k = 0; k < nb; k++ ) {
		color->flags = DoRed | DoGreen | DoBlue;
		color->pixel = p++;
		color->red = i * 65535 / cmap_info->red_max;
		color->green = j * 65535 / cmap_info->green_max;
		color->blue = k * 65535 / cmap_info->blue_max;
	        ++color;
	    }
	}
    }

    XStoreColors( dpy, cmap_info->colormap, colors, num_colors );
    free( colors );
    
    /* Fill in the color approximation information. */
    capx_info->type = PEXColorSpace;
    capx_info->model = PEXColorApproxRGB;
    capx_info->dither = PEXOn;
    capx_info->base_pixel = cmap_info->base_pixel;
    capx_info->max1 = cmap_info->red_max;
    capx_info->max2 = cmap_info->green_max;
    capx_info->max3 = cmap_info->blue_max;
    capx_info->weight1 = 0.0; /* not used by PEXColorSpace */
    capx_info->weight2 = 0.0; /* not used by PEXColorSpace */
    capx_info->weight3 = 0.0; /* not used by PEXColorSpace */
    capx_info->mult1 = cmap_info->red_mult;
    capx_info->mult2 = cmap_info->green_mult;
    capx_info->mult3 = cmap_info->blue_mult;
    
    return 1;
}

int
ora_create_gray_map( dpy, vis_info, num_grays, cmap_info, capx_info )
    Display		*dpy;
    XVisualInfo		*vis_info;
    int			num_grays;
    XStandardColormap	*cmap_info;
    PEXColorApproxEntry	*capx_info;
{
    int			st, i, p;
    unsigned long	*pixels;
    XColor		*colors, *color;
    
    cmap_info->visualid = vis_info->visualid;
    cmap_info->colormap = XCreateColormap( dpy,
	RootWindow( dpy, DefaultScreen( dpy )),
	vis_info->visual, AllocNone );
    cmap_info->red_max = num_grays - 1;
    cmap_info->red_mult = 1;

    pixels = (unsigned long*) malloc( num_grays * sizeof(*pixels) );
    st = XAllocColorCells( dpy, cmap_info->colormap, True,
			  (unsigned long *)NULL, 0, pixels, num_grays );
    if ( st == 0 ) {
	free( pixels );
	return 0;
    }
	
    cmap_info->base_pixel = pixels[0];			  
    free( pixels );

    /* Fill in the RGB color values. */
    colors = (XColor *) malloc( num_grays * sizeof(*colors) );
    p = cmap_info->base_pixel;
    for ( i = 0, color = colors; i < num_grays; i++ ) {
	color->flags = DoRed | DoGreen | DoBlue;
	color->pixel = p++;
	/* R, G, and B are the same intensity within a cell. */
	color->red = color->green = color->blue = 
	    i * 65535 / (num_grays - 1);
	++color;
    }
    
    XStoreColors( dpy, cmap_info->colormap, colors, num_grays );
    free( colors );
    
    /* Fill in the color approximation information. */
    capx_info->type = PEXColorRange;
    capx_info->model = PEXColorApproxRGB;
    capx_info->dither = PEXOn;
    capx_info->base_pixel = cmap_info->base_pixel;
    capx_info->max1 = num_grays - 1;
    capx_info->max2 = 0; /* not used by PEXColorRange */
    capx_info->max3 = 0; /* not used by PEXColorRange */
    /* Give the weights the NTSC intensity coefficients. */
    capx_info->weight1 = 0.299;
    capx_info->weight2 = 0.587;
    capx_info->weight3 = 0.114;
    capx_info->mult1 = 1.0;
    capx_info->mult2 = 0.0;
    capx_info->mult3 = 0.0;

    return 1;
}

int
ora_create_read_only_map( dpy, vis_info, cmap_info, capx_info )
    Display		*dpy;
    XVisualInfo		*vis_info;
    XStandardColormap	*cmap_info;
    PEXColorApproxEntry	*capx_info;
{
    /* Create the colormap. */
    cmap_info->colormap = XCreateColormap( dpy,
	RootWindow( dpy, DefaultScreen( dpy )),
	vis_info->visual, AllocNone );
    
    /* Set up the colormap and color approximation info. */
    cmap_info->base_pixel = 0;
    cmap_info->visualid = vis_info->visualid;
    
    /* The rest depends on the visual class. */
    switch ( vis_info->class ) {
    case TrueColor:
    case StaticColor:
	cmap_info->red_max = vis_info->red_mask;
	cmap_info->red_mult = 1;
	while ( !(cmap_info->red_max & 0x01) ) {
	    cmap_info->red_max >>= 1;
	    cmap_info->red_mult <<= 1;
	}
	cmap_info->green_max = vis_info->green_mask;
	cmap_info->green_mult = 1;
	while ( !(cmap_info->green_max & 0x01) ) {
	    cmap_info->green_max >>= 1;
	    cmap_info->green_mult <<= 1;
	}
	cmap_info->blue_max = vis_info->blue_mask;
	cmap_info->blue_mult = 1;
	while ( !(cmap_info->blue_max & 0x01) ) {
	    cmap_info->blue_max >>= 1;
	    cmap_info->blue_mult <<= 1;
	}

	capx_info->type = PEXColorSpace;
	capx_info->model = PEXColorApproxRGB;
	capx_info->dither = PEXOn;
	capx_info->base_pixel = cmap_info->base_pixel;
	capx_info->max1 = cmap_info->red_max;
	capx_info->max2 = cmap_info->green_max;
	capx_info->max3 = cmap_info->blue_max;
	capx_info->weight1 = 0.0; /* not used by PEXColorSpace */
	capx_info->weight2 = 0.0; /* not used by PEXColorSpace */
	capx_info->weight3 = 0.0; /* not used by PEXColorSpace */
	capx_info->mult1 = cmap_info->red_mult;
	capx_info->mult2 = cmap_info->green_mult;
	capx_info->mult3 = cmap_info->blue_mult;
	break;
    
    case StaticGray:
	cmap_info->red_max = vis_info->colormap_size - 1;
	cmap_info->red_mult = 1;
	
	capx_info->type = PEXColorRange;
	capx_info->model = PEXColorApproxRGB;
	capx_info->dither = PEXOn;
	capx_info->base_pixel = cmap_info->base_pixel;
	capx_info->max1 = cmap_info->red_max;
	capx_info->max2 = 0; /* not used by PEXColorRange */
	capx_info->max3 = 0; /* not used by PEXColorRange */
	/* Give the weights the NTSC intensity coefficients. */
	capx_info->weight1 = 0.299; 
	capx_info->weight2 = 0.587;
	capx_info->weight3 = 0.114;
	capx_info->mult1 = cmap_info->red_mult;
	capx_info->mult2 = 0;
	capx_info->mult3 = 0;
	break;
    }

    return 1;
}    

int double_buffer_supported( dpy, visual, depth )
Display *dpy;
Visual *visual;
int depth;
{
    Window win;
    Multibuffer buffers[5];
    XSetWindowAttributes wattrs;

    wattrs.background_pixel = 0;
    wattrs.border_pixel = 0;
    wattrs.backing_store = NotUseful;
    wattrs.event_mask = ExposureMask;
    wattrs.colormap = XCreateColormap( dpy, 
        RootWindow( dpy, DefaultScreen(dpy) ), visual, AllocNone );
    win = XCreateWindow( dpy, RootWindow( dpy, DefaultScreen(dpy) ),
        0, 0, 10, 10, 1, depth, InputOutput, visual, 
        CWBackPixel | CWBorderPixel | CWBackingStore | CWEventMask | CWColormap,

        &wattrs );

    if( ora_create_mbx_buffers( dpy, win, 2, buffers ) )
      return 1;
    else
      return 0;
}

int
ora_create_mbx_buffers( dpy, window, num_buffers, buffers )
Display    *dpy;
Window    window;
int        num_buffers;
Multibuffer    *buffers;
{
    int        first_event, first_error, num;

    if ( !XmbufQueryExtension( dpy, &first_event, &first_error ) ) {
        if( verbose )
        fprintf( stderr, "Utils:\tNo multi-buffer extension on %s\n",
            XDisplayString( dpy ) );
        return 0;
    }

    num = XmbufCreateBuffers( dpy, window, num_buffers,
        MultibufferUpdateActionUndefined,
        MultibufferUpdateHintFrequent,
        buffers );
    if ( num != num_buffers ) {
        fprintf( stderr, "Utils: Can't create %d buffers, only %d\n",
            num_buffers, num );
        if ( num > 0 )
            XmbufDestroyBuffers( dpy, window );
        return 0;
    }

    return num;
}
